#version 300 es

precision mediump sampler2DArray;

#define GLFX_IBL
#define GLFX_TBN
#define GLFX_TEX_MRAO
#define GLFX_LIGHTING
#define GLFX_2SIDED

layout( location = 0 ) in vec3 attrib_pos;
#ifdef GLFX_LIGHTING
layout( location = 1 ) in vec3 attrib_n;
#ifdef GLFX_TBN
layout( location = 2 ) in vec4 attrib_t;
#endif
#endif
layout( location = 3 ) in vec2 attrib_uv;
layout( location = 4 ) in uvec4 attrib_bones;
#ifndef GLFX_1_BONE
layout( location = 5 ) in vec4 attrib_weights;
#endif

layout(std140) uniform glfx_GLOBAL
{
	mat4 glfx_MVP;
	mat4 glfx_PROJ;
	mat4 glfx_MV;
	vec4 glfx_QUAT;

	vec4 js_scale;
	vec4 js_rotation;
};
layout(std140) uniform glfx_INSTANCES
{
    vec4 glfx_IDATA[48];
};
uniform uint glfx_CURRENT_I;
#define glfx_T_SPAWN (glfx_IDATA[glfx_CURRENT_I].x)
#define glfx_T_ANIM (glfx_IDATA[glfx_CURRENT_I].y)
#define glfx_ANIMKEY (glfx_IDATA[glfx_CURRENT_I].z)

uniform sampler2D glfx_BONES;

#ifdef GLFX_USE_UVMORPH
uniform sampler2DArray glfx_UVMORPH_FISHEYE;
#endif

#ifdef GLFX_OCCLUSION
layout(std140) uniform glfx_OCCLUSION_DATA
{
    vec4 glfx_OCCLUSION_RECT;
};
out vec2 glfx_OCCLUSION_UV;
#endif

#ifdef GLFX_USE_AUTOMORPH
uniform sampler2D glfx_MORPH;
vec2 glfx_morph_coord( vec3 v )
{
    const float half_angle = radians(104.);
    const float y0 = -110.;
    const float y1 = 112.;
    float x = atan( v.x, v.z )/half_angle;
    float y = ((v.y-y0)/(y1-y0))*2. - 1.;
    return vec2(x,y);
}
#ifndef GLFX_AUTOMORPH_BONE
vec3 glfx_auto_morph( vec3 v )
{
    vec2 morph_uv = glfx_morph_coord(v)*0.5 + 0.5;
    vec3 translation = texture( glfx_MORPH, morph_uv ).xyz;
    return v + translation;
}
#else
vec3 glfx_auto_morph_bone( vec3 v, mat3x4 m )
{
    vec2 morph_uv = glfx_morph_coord(vec3(m[0][3],m[1][3],m[2][3]))*0.5 + 0.5;
    vec3 translation = texture( glfx_MORPH, morph_uv ).xyz;
    return v + translation;
}
#endif
#endif

#ifdef GLFX_USE_SHADOW
out vec3 var_shadow_coord;
vec3 spherical_proj( vec2 fovM, vec2 fovP, float zn, float zf, vec3 v )
{
    vec2 xy = (atan( v.xy, v.zz )-(fovP+fovM)*0.5)/((fovP-fovM)*0.5);
    float z = (length(v)-(zn+zf)*0.5)/((zf-zn)*0.5);
    return vec3( xy, z );
}
#endif

out vec2 var_uv;
#ifdef GLFX_LIGHTING
#ifdef GLFX_TBN
out vec3 var_t;
out vec3 var_b;
#endif
out vec3 var_n;
out vec3 var_v;
#endif

mat3x4 get_bone( uint bone_idx, float k )
{
	float bx = float( int(bone_idx)*3 );
	vec2 rts = 1./vec2(textureSize(glfx_BONES,0));
	return mat3x4( 
		texture( glfx_BONES, (vec2(bx,k)+0.5)*rts ),
		texture( glfx_BONES, (vec2(bx+1.,k)+0.5)*rts ),
		texture( glfx_BONES, (vec2(bx+2.,k)+0.5)*rts ) );
}

void main()
{
    mat3x4 m = get_bone( attrib_bones[0], glfx_ANIMKEY );

	vec3 vpos = attrib_pos;
	vpos = vec4(vpos+vec3(0.,0.,-16), 1.) * m;
	vec3 translation = vec3(glfx_MV[3].xyz);
	vpos.y = 1. - vpos.y;
	vec3 scale;
	scale.x = js_scale.x;
	scale.y = js_scale.x;
	scale.z = js_scale.x;

	mat4 rotationMatrix = mat4(
		glfx_MV[0][0] / scale.x, glfx_MV[0][1] / scale.x, glfx_MV[0][2] / scale.x, 0.,
		glfx_MV[1][0] / scale.y, glfx_MV[1][1] / scale.y, glfx_MV[1][2] / scale.y, 0.,
		glfx_MV[2][0] / scale.z, glfx_MV[2][1] / scale.z, glfx_MV[2][2] / scale.z, 0.,
		0., 0., 0., 1.
	);

	// float x = atan(-rotationMatrix[1][2], rotationMatrix[2][2]);
	float cosY = sqrt(1. - rotationMatrix[0][2]);
	// float y = atan(rotationMatrix[0][2], cosY);
	
	float cosX = -cos(radians(js_rotation.x));
	float sinX = sin(radians(js_rotation.x));

	// float sinY = sin(radians(js_rotation.x));
	// float cosY = cos(radians(js_rotation.y));

	float sinZ = cosX * rotationMatrix[1][0] + sinX * rotationMatrix[2][0];
	float cosZ = cosX * rotationMatrix[1][1] + sinX * rotationMatrix[2][1];
 	float z = atan(sinZ, cosZ);

	// mat4 rot_x = mat4(1., 0., 0., 0.,
	// 									0., cosX, -sinX, 0.,
	// 									0., sinX, cosX, 0.,
	// 									0., 0., 0., 1.);

	// mat4 rot_y = mat4(cos, 0, sin, 0.,
	// 								0,  1., 0., 0.,
	// 								-sinY, 0., cos, 0.,
	// 								0., 0., 0., 1.);

	mat4 rot_z = mat4(cosX, -sinX, 0., 0.,
										sinX, cosX, 0., 0.,
										0., 0., 1., 0.,
										0., 0., 0., 1.);
	
	mat4 scale_matrix = mat4(scale.x, 0., 0., 0.,
														0., scale.y, 0., 0.,
														0., 0., scale.z, 0.,
														0., 0., 0., 1.);

	mat4 translation_matrix = mat4(1., 0., 0., 0.,
																 0., 1., 0., 0.,
																 0., 0., 1., 0.,
																 translation.x, translation.y, translation.z, 1.);


	mat4 new_transform = translation_matrix * scale_matrix * rot_z;
	
	gl_Position = glfx_PROJ * new_transform * vec4(vpos,1.);
	// gl_Position.z = -1.;

	var_uv = attrib_uv;

#ifdef GLFX_LIGHTING
    var_n = mat3(glfx_MV)*(attrib_n*mat3(m));
#ifdef GLFX_TBN
    var_t = mat3(glfx_MV)*(attrib_t.xyz*mat3(m));
    var_b = attrib_t.w*cross( var_n, var_t );
#endif
    var_v = (glfx_MV*vec4(vpos,1.)).xyz;
#endif

}